#include <conio.h>
#include <dos.h>
// The following two are necessary for the random generator
#include <stdlib.h>
#include <time.h>

char far *screen;

/**************************************************************************
** Set to VGA 320x200 256 colours                                        **
**************************************************************************/
void VGAScreen ()
{
  asm {
    mov ax, 0x13
    int 0x10
  }
}

/**************************************************************************
** Set to text mode                                                      **
**************************************************************************/
void TextScreen ()
{
  asm {
    mov ax, 0x3
    int 0x10
  }
}

/**************************************************************************
** Sets a DAC register to a specific Red Green Blue-value                **
**************************************************************************/
void SetDAC (unsigned char DAC, unsigned char R, unsigned char G, unsigned char B)
{
  outportb (0x3c8, DAC);
  outportb (0x3c9, R);
  outportb (0x3c9, G);
  outportb (0x3c9, B);
}

/**************************************************************************
** Sets up a smooth palette ranging from blue to white                   **
** This leaves colour 0 untouched                                        **
**************************************************************************/
void SmoothPalette ()
{
  unsigned char Value;

  for (Value = 1; Value < 255; Value++) {
    SetDAC(Value,
           Value >> 2, //Red fades in from none to full
           Value >> 2, //Green fades in from none to full
           63);        //Blue is always at full strength
  }
}

/**************************************************************************
** Sets a single pixel on the screen to a specified colour               **
**************************************************************************/
void PutPixel (unsigned int x, unsigned int y, unsigned char c)
{
  screen[(y << 8) + (y << 6) + x] = c;
}

/**************************************************************************
** Reads the colour of a specified pixel on the screen                   **
**************************************************************************/
unsigned char ReadPixel (unsigned int x, unsigned int y)
{
  return (screen[(y << 8) + (y << 6) + x]);
}


void Plasma (unsigned int x1, unsigned int y1, unsigned int x2, unsigned int y2)
{
  unsigned char
    TopLeft,
    TopRight,
    BottomLeft,
    BottomRight,
    Left,
    Right,
    Top,
    Bottom,
    Center;

  //Make sure recursion ends at a point
  if (x2 - x1 < 2 && y2 - y1 < 2) return;
  //Read colours
  TopLeft     = ReadPixel (x1, y1);
  TopRight    = ReadPixel (x2, y1);
  BottomLeft  = ReadPixel (x1, y2);
  BottomRight = ReadPixel (x2, y2);
  //Average colours
  Left   = (TopLeft + BottomLeft) >> 1;
  Right  = (TopRight + BottomRight) >> 1;
  Top    = (TopLeft + TopRight) >> 1;
  Bottom = (BottomLeft + BottomRight) >> 1;
  Center = (TopLeft + TopRight + BottomLeft + BottomRight) >> 2;
  //Define four quadrants
  PutPixel (x1, (y1 + y2) >> 1, Left);
  PutPixel (x2, (y1 + y2) >> 1, Right);
  PutPixel ((x1 + x2) >> 1, y1, Top);
  PutPixel ((x1 + x2) >> 1, y2, Bottom);
  PutPixel ((x1 + x2) >> 1, (y1 + y2) >> 1, Center);
  //Make a smaller plasma in each of the four quadrants
  Plasma (x1, y1, (x1 + x2) >> 1, (y1 + y2) >> 1); //top left
  Plasma ((x1 + x2) >> 1, (y1 + y2) >> 1, x2, y2); //bottom right
  Plasma ((x1 + x2) >> 1, y1, x2, (y1 + y2) >> 1); //top right
  Plasma (x1, (y1 + y2) >> 1, (x1 + x2) >> 1, y2); //bottom left
  //Ah! Recursion
}

void main ()
{
  unsigned char p, q;

  //Set up a pointer to the vga memory at A000:0000
  screen = MK_FP (0xA000,0);

  //Clear Screen
  clrscr ();

  //Display message
  printf ("PLASMA Demonstration by Steven Don\n");
  printf ("----------------------------------\n");
  printf ("This will generate a BIG plasma cloud.\n");
  printf ("It has only four different colours, one at each corner.\n");
  printf ("Because of this, it looks far too smooth.\n");
  printf ("\n");
  printf ("Press any key to see what I mean.\n");
  //Wait for a keypress
  getch ();

  //Switch to VGA mode
  VGAScreen ();

  //Set up palette to shades of blue
  SmoothPalette ();

  //Make sure we get a different plasma cloud each time
  randomize ();

  //Create one large plasma, too smooth, but you can see how it works
  PutPixel (  0,   0, random (254) + 1);
  PutPixel (319,   0, random (254) + 1);
  PutPixel (  0, 199, random (254) + 1);
  PutPixel (319, 199, random (254) + 1);
  Plasma (0, 0, 319, 199);
  //Wait for a keypress
  getch ();

  //Switch back to text
  TextScreen ();

  //Display message
  printf ("PLASMA Demonstration by Steven Don\n");
  printf ("----------------------------------\n");
  printf ("However, if you'd make the big one a bit smaller and combined more of them,\n");
  printf ("it would look a lot better.\n");
  printf ("\n");
  printf ("Press any key to see the result.\n");
  //Wait for a keypress
  getch ();

  //Switch back to VGA
  VGAScreen ();
  //Palette was changed when we switched to text, so set it again
  SmoothPalette ();

  //Create 16 small plasma's, much better
  for (p = 0; p <= 3; p++) {
    //Set top left pixels
    PutPixel ( 80 + p * 40, 20, random (254) + 1);
    PutPixel ( 80, 20 + p * 40, random (254) + 1);
  }
  for (q = 0; q <= 3; q++)
    for (p = 0; p <= 3; p++) {
      //Set bottom right pixel
      PutPixel (120 + p * 40, 60 + q * 40, random (254) + 1);
      //Generate small plasma
      Plasma ( 80 + p * 40, 20 + q * 40, 120 + p * 40, 60 + q * 40);
    }

  //Wait for a keypress
  getch ();

  //Switch back to text
  TextScreen ();
}